/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __operator_SurfaceWrapOp_h__
#define __operator_SurfaceWrapOp_h__

namespace fe
{

/**************************************************************************//**
	@brief Operator to update an attached surface to follow a driving surface

	@ingroup operator
*//***************************************************************************/
class FE_DL_EXPORT SurfaceWrapOp:
	public SurfaceBindOp,
	public Initialize<SurfaceWrapOp>,
	public ObjectSafe<SurfaceWrapOp>
{
	public:

					SurfaceWrapOp(void):
							m_brushed(FALSE),
						m_lastFrame(-1.0),
						m_forceRecook(FALSE),
						m_selectIndex(-1)									{}
virtual				~SurfaceWrapOp(void)									{}

		void		initialize(void);

					//* As HandlerI
virtual	void		handle(Record& a_rSignal);

virtual	void		run(I32 a_id,sp<SurfaceRange> a_spRange);

	class Feedback
	{
		public:
									Feedback(Real a_weight,
										sp<SurfaceI::ImpactI> a_spImpact):
										m_weight(a_weight),
										m_spImpact(a_spImpact)				{}

			Real					weight(void) const	{ return m_weight; }
			sp<SurfaceI::ImpactI>	impact(void) const	{ return m_spImpact; }

		private:
			Real					m_weight;
			sp<SurfaceI::ImpactI>	m_spImpact;
	};

	private:

		sp<SurfaceAccessibleI>					m_spInputAccessible;
		sp<SurfaceAccessorI>					m_spFragmentAccessor;

		sp<SurfaceI>							m_spDeformed;
		sp<SurfaceI>							m_spOutput;

		sp<DrawI>								m_spDrawDebug;
		sp<DrawMode>							m_spDebugGroup;
		sp<DrawMode>							m_spOverlay;
		sp<DrawMode>							m_spBrush;

		BWORD									m_brushed;
		Real									m_lastFrame;
		I32										m_forceRecook;
		WindowEvent								m_event;
		I32										m_selectIndex;
		std::vector<Feedback>					m_feedback;

		I32										m_bindCount;
};

} /* namespace */

#endif /* __operator_SurfaceWrapOp_h__ */
