/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __surface_SurfaceAccessibleBase_h__
#define __surface_SurfaceAccessibleBase_h__

namespace fe
{

/**************************************************************************//**
    @brief Common Functionality for Accessible Surface

	@ingroup surface
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessibleBase:
	public ObjectSafeShared<SurfaceAccessibleBase>,
	virtual public SurfaceAccessibleI
{
	public:
								SurfaceAccessibleBase(void):
									m_paging(FALSE)							{}

virtual							~SurfaceAccessibleBase(void)				{}

								//* As Protectable
virtual	Protectable*			clone(Protectable* pInstance=NULL);

								//* as SurfaceAccessibleI

virtual	void					bind(Instance a_instance)
								{
									Record record=a_instance.cast<Record>();
									bind(record);
									if(record.isValid())
									{
										sp<Scope> spScope=
												record.layout()->scope();
										bind(spScope);
									}
								}
virtual	BWORD					isBound(void)	{ return m_record.isValid(); }

virtual	void					bind(sp<Scope> a_spScope)
								{	m_spScope=a_spScope; }
virtual	BWORD					load(String a_filename);
virtual	BWORD					save(String a_filename);

virtual	void					bind(sp<SurfaceAccessibleI::ThreadingState>
										a_spThreadingState)
								{	m_spThreadingState=a_spThreadingState; }
virtual	void					lock(void)		{ safeLock(); }
virtual	void					unlock(void)	{ safeUnlock(); }

virtual
const	std::vector<SurfaceAccessibleI::Spec>&
								attributeSpecs(SurfaceAccessibleI::Element
										a_element) const
								{	return m_specs; }

virtual	void					copy(sp<SurfaceAccessibleI>
										a_spSurfaceAccessible);

virtual void					require(SurfaceAccessibleI::Element a_element,
										String a_name,String a_typeName);

virtual	I32						count(SurfaceAccessibleI::Element
										a_element) const;

virtual sp<SurfaceAccessorI>	accessor(Element a_element,String a_name,
										Creation a_create);

virtual sp<SurfaceAccessorI>	accessor(Element a_element,
										Attribute a_attribute,
										Creation a_create);

virtual sp<SurfaceAccessorI>	accessor(Element a_element,String a_name)
								{	return accessor(a_element,a_name,
											e_createMissing); }

virtual sp<SurfaceAccessorI>	accessor(Element a_element,
										Attribute a_attribute)
								{	return accessor(a_element,a_attribute,
											e_createMissing); }

virtual	BWORD					discard(SurfaceAccessibleI::Element a_element,
										String a_name)
								{	return FALSE; }
virtual	BWORD					discard(SurfaceAccessibleI::Element a_element,
										SurfaceAccessibleI::Attribute
										a_attribute);

virtual	sp<SurfaceI>			surface(void)
								{	return surface(String()); }
virtual	sp<SurfaceI>			surface(String a_group);
virtual	sp<SurfaceI>			subSurface(U32 a_subIndex)
								{	return subSurface(a_subIndex,String()); }
virtual	sp<SurfaceI>			subSurface(U32 a_subIndex,String a_group);

virtual	void					append(sp<SurfaceAccessibleI>
										a_spSurfaceAccessibleI,
										const SpatialTransform&
										a_rTransform)						{}

virtual	void					instance(sp<SurfaceAccessibleI>
										a_spSurfaceAccessibleI,
										const std::vector<SpatialTransform>&
										a_rTransformArray)					{}

								//* default implementation is non-atomic
virtual	sp<SurfaceRange>		atomize(AtomicChange a_atomicChange,
										String a_group,U32 a_desiredCount);

virtual void					setPaging(BWORD a_paging)
								{	m_paging=a_paging; }
virtual BWORD					paging(void) const	{ return m_paging; }

virtual void					preparePaging(AtomicChange a_atomicChange,
										String a_group)						{}
virtual I32						pointPage(U32 a_pointIndex) const
								{	return 0; }
virtual I32						primitivePage(U32 a_primitiveIndex) const
								{	return 0; }

		void					bind(Record& a_rRecord)	{ m_record=a_rRecord; }
		Record					record(void)			{ return m_record; }

	protected:
		String					elementLayout(
										SurfaceAccessibleI::Element a_element);

		sp<SurfaceRange>		atomizeConnectivity(AtomicChange a_atomicChange,
										String a_group,U32 a_desiredCount);

		SurfaceAccessibleI::Threading	threading(void) const
							{
								if(m_spThreadingState.isNull())
								{
									return SurfaceAccessibleI::e_unknown;
								}
								return Threading(
										m_spThreadingState->threading());
							}

		sp<SurfaceAccessibleI::ThreadingState>	m_spThreadingState;

		sp<Scope>				m_spScope;
		sp<RecordGroup>			m_spRecordGroup;
		WeakRecord				m_record;
		BWORD					m_paging;

mutable	std::vector<Spec>		m_specs;
};

} // namespace

#endif /* __surface_SurfaceAccessibleBase_h__ */
